import processing.core.*; 
import processing.xml.*; 

import fisica.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class Sumogons extends PApplet {

/**
 *  Polygons
 *
 *  by Ricard Marxer
 *
 *  This example shows how to create polygon bodies.
 */



Stack states;
FWorld world;
PlayerController[] players;
int[] leaderboard;
int[] playerSides;
ArrayList<PolygonController> shapes;
ArrayList<EnemyController> enemies;
PFont omnesSem20;
PFont omnesSem29;
PFont omnesSem48;

int numPlayers = 2;
int maxScore = 3;

int numSides = 3;
int minSides = 3;
PVector scale; //Scale of the player & newly spawned debug objects
PVector up;
PVector down;
PVector left;
PVector right;
PVector centerOfScreen;
float scaleAdjustFactor = 0.3f; //0.1 for mouse
float movementForce = 1000000;
float interstitialLengthMs = 1000;

//Re-using some PVector objects to reduce garbage during calcs in a tight loop
PVector gWorkVectorA;

static final boolean DEBUG = false;

boolean disableEnemyUpdate = false;

public void setup() {
  size(1024, 768);
  smooth();
  
  states = new Stack();
  shapes = new ArrayList<PolygonController>();
  enemies = new ArrayList<EnemyController>();
  scale = new PVector(1,1);
  up = new PVector(0, 1);
  down = new PVector(0, -1);
  left = new PVector(1, 0);
  right = new PVector(-1, 0);
  centerOfScreen = new PVector(width / 2, height / 2);
  players = new PlayerController[numPlayers];
  playerSides = new int[numPlayers];
  gWorkVectorA = new PVector();
  omnesSem20  = loadFont("Omnes-Semibold-20.vlw");
  omnesSem29  = loadFont("Omnes-Semibold-29.vlw");
  omnesSem48  = loadFont("Omnes-Semibold-48.vlw");

  /* //Not using mouse ATM
  addMouseWheelListener(new java.awt.event.MouseWheelListener() { 
    public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt) { 
      mouseWheel(evt.getWheelRotation());
  }});
  */

  Fisica.init(this);

  //Begin initial game state
  engineChangeState(new IntroState());
}

/**
 * Listener for Mouse wheel movement
 */
/* //Not using mouse ATM
void mouseWheel(int delta) {
  //println(delta); 
  PlayerController player = players[0];
  Input input = player.getInput();
  //if (input.scaleModKeyDown) {
  //  scale.y += delta * scaleAdjustFactor;
  //} else {
  //  scale.x += delta * scaleAdjustFactor;
  //}
  scale.x += delta * scaleAdjustFactor;
  scale.y += delta * scaleAdjustFactor;
  player.setScale(scale.x, scale.y);
  player.updateShape();
}
*/


public void mouseDragged() {
  engineGetState().mouseDragged();
}

public void mousePressed() {
  engineGetState().mousePressed();
}

public void mouseReleased() {
  engineGetState().mouseReleased();
}

public void keyPressed() {
  engineGetState().keyPressed();
}

public void keyReleased() {
  engineGetState().keyReleased();
}

public PApplet getMainInstance () {
  //Stupid hack because some libraries require a reference to the main PApplet
  return this;
}

public EnemyController findEnemy(FBody b) {
  for (EnemyController enemy : enemies) {
    if (enemy.getPoly() == b) {
      return enemy;
    }
  }
  return null;
}

public void draw() {
  engineGetState().update();
  engineGetState().draw();
}

public GameState engineGetState() {
  if (!states.isEmpty()) {
    return (GameState)states.peek(); 
  } else {
    return null;
  }
}

public void engineChangeState(GameState state) {
  //Cleanup current state
  if (!states.isEmpty()) {
    GameState currentState = (GameState)states.peek();
    currentState.cleanup();
    states.pop();
  }

  //Store and setup new state
  states.push(state);
  state.setup();
}

public void enginePushState(GameState state) {
  //Cleanup current state
  if (!states.isEmpty()) {
    GameState currentState = (GameState)states.peek();
    currentState.pause();
  }

  //Store and setup new state
  states.push(state);
  state.setup();
}

public void enginePopState() {
  GameState previousState = null;
  //Cleanup current state
  if (!states.isEmpty()) {
    GameState currentState = (GameState)states.peek();
    currentState.cleanup();
    previousState = currentState;
    states.pop();
  }

  //Resume previous state
  if (!states.isEmpty()) {
    GameState currentState = (GameState)states.peek();
    currentState.resume(previousState);
  }
}

/**
 * Clear the state stack in reverse order, 
 */
public void engineResetToState(GameState state) {
  //Empty out current state stack, giving each state a chance to run cleanup()
  if (!states.isEmpty()) {
    for (int i = states.size(); i > 0;) {
      GameState currentState = (GameState)states.peek();
      currentState.cleanup();
      states.pop();
      i = states.size();
    }
  }

  engineChangeState(state);
}
class EnemyController extends PolygonController {
  float patrolRadius = 5;

  Path path;
  int nextPathNodeIndex = 0;
  static final float targetNodeDistance = 150.0f;
  float enemyMovementForce;
  boolean attackPlayer = true; //Pursue the player?

  EnemyController(int numSides, float radius) {
    super(numSides, radius);
    construct();
  }
  
  EnemyController(int numSides, float radius, FWorld world) {
    super(numSides, radius, world);
    construct();
  }

  public void construct() {
    invincible = false;
    enemyMovementForce = movementForce / 350;
    path = new Path();
    turnSpeed = 4;
    //Auto-follow path for enemy
    path.add(new PVector(0, height/2));
    path.add(new PVector(width/2, height));
    path.add(new PVector(width, height/2));
    path.add(new PVector(width/2, 0));
  }

  public void update() {
    super.update();

    if (null != poly) {
      PVector pos = getPosition();

      //Patrol
      if (path.size() > 0) {
        PVector nextNode = path.get(nextPathNodeIndex);
        if (null != nextNode) {
          //Figure out where we're going
          PVector moveVector = PVector.sub(nextNode, pos);
          float distanceToNode = moveVector.mag();

          //Amplify & apply movement force
          moveVector.mult(enemyMovementForce);
          setForce(moveVector.x, moveVector.y);

          //If we're close enough to this node, switch to the next one.
          if (distanceToNode <= targetNodeDistance) {
            nextPathNodeIndex++;
            if (!(nextPathNodeIndex < path.size())) {
              //Loop back to beginning of path
              nextPathNodeIndex = 0;
            }
          }
        }
      }

      PVector lookAt = PVector.sub(pos, centerOfScreen); //FIXME: Use workvector placeholders if this creates too much garbage
      lookAt.normalize();
      getHeading(); //update this.heading

      float angleBetween = -PVector.dot(lookAt, heading);

      poly.setAngularVelocity(angleBetween * turnSpeed);
    }

  }

  public void drawPath() {
    path.draw();
  }
}
/**
 * Taken from the Spectrum Game Engine
 *
 * @author Eliot Lash
 * @copyright Copyright (c) 2010-2012 Eliot Lash
 */
class GameState {
  //'Virtual' methods to be overriden in subclasses
  public void setup() {
  }

  public void cleanup() {
  }

  public void pause() {
  }

  public void resume(GameState previousState) {
  }

  public void update() {
  }

  public void draw() {
  }

  public void mouseDragged() {
  }

  public void mousePressed() {
  }

  public void mouseReleased() {
  }

  public void keyPressed() {
  }

  public void keyReleased() {
  }

}
class Input {
  boolean useMouse = false;

  boolean centerKeyDown = false;
  boolean upKeyDown = false;
  boolean downKeyDown = false;
  boolean leftKeyDown = false;
  boolean rightKeyDown = false;
  boolean scaleXNegativeKeyDown = false;
  boolean scaleXPositiveKeyDown = false;
  boolean scaleYNegativeKeyDown = false;
  boolean scaleYPositiveKeyDown = false;
  boolean scaleModKeyDown = false;

  PVector heading;
  float rotation;

  Key CENTER_KEY;
  Key UP_KEY;
  Key DOWN_KEY;
  Key LEFT_KEY;
  Key RIGHT_KEY;
  Key SCALE_MOD_KEY;
  Key SCALE_X_POSITIVE_KEY;
  Key SCALE_X_NEGATIVE_KEY;
  Key SCALE_Y_POSITIVE_KEY;
  Key SCALE_Y_NEGATIVE_KEY;

  Input() {
    heading = up.get();
    CENTER_KEY = new Key();
    UP_KEY = new Key();
    DOWN_KEY = new Key();
    LEFT_KEY = new Key();
    RIGHT_KEY = new Key();
    SCALE_MOD_KEY = new Key();
    SCALE_X_POSITIVE_KEY = new Key();
    SCALE_X_NEGATIVE_KEY = new Key();
    SCALE_Y_POSITIVE_KEY = new Key();
    SCALE_Y_NEGATIVE_KEY = new Key();
  }

  Input(Input otherInput) {
    scaleModKeyDown = otherInput.scaleModKeyDown;
    upKeyDown = otherInput.upKeyDown;
    downKeyDown = otherInput.downKeyDown;
    leftKeyDown = otherInput.leftKeyDown;
    rightKeyDown = otherInput.rightKeyDown;
    scaleXNegativeKeyDown = otherInput.scaleXNegativeKeyDown;
    scaleXPositiveKeyDown = otherInput.scaleXPositiveKeyDown;
    scaleYNegativeKeyDown = otherInput.scaleYNegativeKeyDown;
    scaleYPositiveKeyDown = otherInput.scaleYPositiveKeyDown;

    CENTER_KEY = otherInput.CENTER_KEY;
    UP_KEY = otherInput.UP_KEY;
    DOWN_KEY = otherInput.DOWN_KEY;
    LEFT_KEY = otherInput.LEFT_KEY;
    RIGHT_KEY = otherInput.RIGHT_KEY;
    SCALE_MOD_KEY = otherInput.SCALE_MOD_KEY;
    SCALE_X_POSITIVE_KEY = otherInput.SCALE_X_POSITIVE_KEY;
    SCALE_X_NEGATIVE_KEY = otherInput.SCALE_X_NEGATIVE_KEY;
    SCALE_Y_POSITIVE_KEY = otherInput.SCALE_Y_POSITIVE_KEY;
    SCALE_Y_NEGATIVE_KEY = otherInput.SCALE_Y_NEGATIVE_KEY;
  }

  public void keyPressed(Key k) {
    setKeyDown(k, true);
  }

  public void keyReleased(Key k) {
    setKeyDown(k, false);
  }

  private void setKeyDown(Key k, boolean down) {
    if (CENTER_KEY.equals(k)) {
      centerKeyDown = down;
    } else if (UP_KEY.equals(k)) {
      upKeyDown = down;
    } else if (DOWN_KEY.equals(k)) {
      downKeyDown = down;
    } else if (LEFT_KEY.equals(k)) {
      leftKeyDown = down;
    } else if (RIGHT_KEY.equals(k)) {
      rightKeyDown = down;
    } else if (SCALE_MOD_KEY.equals(k)) {
      scaleModKeyDown = down;
    } else if (SCALE_X_NEGATIVE_KEY.equals(k)) {
      scaleXPositiveKeyDown = down;
    } else if (SCALE_X_POSITIVE_KEY.equals(k)) {
      scaleXNegativeKeyDown = down;
    } else if (SCALE_Y_NEGATIVE_KEY.equals(k)) {
      scaleYPositiveKeyDown = down;
    } else if (SCALE_Y_POSITIVE_KEY.equals(k)) {
      scaleYNegativeKeyDown = down;
    }
  }
}
/**
 * Taken from the Spectrum Game Engine
 *
 * @author Eliot Lash
 * @copyright Copyright (c) 2010-2012 Eliot Lash
 */
class InstructionState extends GameState {
  String message;
  GameState nextState;
  boolean resetStateStack = false;
  int backgroundColor = color(66);
  int verticalOffset = -50;

  InstructionState() {
    nextState = new NewGameState();
  }

  public void setup() {
  }

  public void draw() {
    smooth();
    background(backgroundColor);
    textAlign(CENTER);
    fill(255);
    textFont(omnesSem48);
    text("SumoGONS!", width/2, 200); 
    textFont(omnesSem29);
    text("Knock the other player's center off the screen 3 times to win.", width/2, (height/2 + verticalOffset)); 
    text("Red moves with arrow keys, changes shape with IJKL keys.", width/2, (height/2 + verticalOffset + 100)); 
    text("Blue moves with WASD keys, changes shape with TFGH keys.", width/2, (height/2 + verticalOffset + 200)); 
    text("Press space to start.", width/2, (height/2 + verticalOffset + 300)); 
  }

  public void mousePressed() {
  }

  public void goToNextState() {
    if (resetStateStack) {
      engineResetToState(nextState);
    } else {
      engineChangeState(nextState);
    }
  }

  public void keyPressed() {
    if ((ENTER == key) ||
        (RETURN == key) ||
        (' ' == key)
      ) {
      goToNextState();
    }
  }
}
/**
 * Taken from the Spectrum Game Engine
 *
 * @author Eliot Lash
 * @copyright Copyright (c) 2010-2012 Eliot Lash
 */
class InterstitialState extends GameState {
  String message;
  GameState nextState;
  boolean resetStateStack = false;
  int backgroundColor = color(66);
  int verticalOffset = -50;
  float displayLengthMs = -1;
  float displayStartMs = 0;

  InterstitialState(String _message, GameState _nextState) {
    super();
    message = _message;
    nextState = _nextState;
  }

  InterstitialState(String _message, GameState _nextState, float _displayLengthMs) {
    super();
    message = _message;
    nextState = _nextState;
    displayLengthMs = _displayLengthMs;
  }

  InterstitialState(String _message, GameState _nextState, int _verticalOffset) {
    super();
    message = _message;
    nextState = _nextState;
    verticalOffset = _verticalOffset;
  }

  InterstitialState(String _message, GameState _nextState, boolean _resetStateStack) {
    super();
    message = _message;
    nextState = _nextState;
    resetStateStack = _resetStateStack;
  }

  InterstitialState(String _message, GameState _nextState, boolean _resetStateStack, int _backgroundColor) {
    super();
    message = _message;
    nextState = _nextState;
    resetStateStack = _resetStateStack;
    backgroundColor = _backgroundColor;
  }

  public void setup() {
  }

  public void update() {
    if (displayLengthMs >= 0) {
      if (displayStartMs == 0) {
        displayStartMs = millis(); //Set timer
      } else {
        if((millis() - displayStartMs) > displayLengthMs) {
          //Timer's up, close this message
          goToNextState();
        }
      }
    }
  }

  public void draw() {
    smooth();
    background(backgroundColor);
    textFont(omnesSem48);
    textAlign(CENTER);
    fill(255);
    text(message, width/2, (height/2 + verticalOffset)); 
  }

  public void mousePressed() {
  }

  public void goToNextState() {
    if (resetStateStack) {
      engineResetToState(nextState);
    } else {
      engineChangeState(nextState);
    }
  }

  public void keyPressed() {
    if ((ENTER == key) ||
        (RETURN == key) ||
        (' ' == key)
      ) {
      goToNextState();
    }
  }
}
/**
 * Taken from the Spectrum Game Engine
 *
 * @author Eliot Lash
 * @copyright Copyright (c) 2010-2012 Eliot Lash
 */
class IntroState extends GameState {
  //'Virtual' methods to be overriden in subclasses
  public void setup() {
    //Make the world
    world = new FWorld();
    world.setGrabbable(false);
    world.setGravity(0, 0);
    world.setEdges();
    world.setEdgesRestitution(0.5f);

    //Make NPCs for the title screen
    {
      EnemyController enemy = new EnemyController(3, 58.0f, world);
      enemy.setPosition(width/3, height/3);
      enemy.setFill(new PVector(255, 0, 0));
      enemy.updateShape();
      enemies.add(enemy);
    }

    {
      EnemyController enemy = new EnemyController(5, 50.0f, world);
      enemy.setPosition(width/3 + 100, height/3);
      enemy.setFill(new PVector(0, 0, 255));
      enemy.updateShape();
      enemies.add(enemy);
    }
  }

  public void cleanup() {
  }

  public void pause() {
  }

  public void resume(GameState previousState) {
  }

  public void update() {
    for (EnemyController enemy : enemies) {
      enemy.update();
    }

    world.step();
  }

  public void draw() {
    background(99);

    if (DEBUG) {
      for (EnemyController enemy : enemies) {
        enemy.drawPath();
      }
    }

    world.draw();

    textAlign(CENTER);
    fill(255);
    textFont(omnesSem48);
    text("SumoGONS!", width/2, 200); 
    textFont(omnesSem29);
    text("by Eliot Lash\nfor GDC Pirate Kart 2012", width/2, 300); 
    text("Press Space", width/2, (height/2 + 300)); 
  }

  public void mouseDragged() {
  }

  public void mousePressed() {
  }

  public void mouseReleased() {
  }

  public void keyPressed() {
  }

  public void keyReleased() {
    if ((ENTER == key) ||
        (RETURN == key) ||
        (' ' == key)
      ) {
      engineChangeState(new InstructionState());
    }
  }

}
class Key {
  boolean coded = false;
  char myKey = 0;
  int myKeyCode = 0; 

  Key() {
  }

  Key(char _key, int _keyCode) {
    coded = (CODED == key);
    myKey = _key;
    myKeyCode = _keyCode;
  }

  Key(char _key) {
    myKey = _key; 
    coded = false;
  }

  Key(int _keyCode, boolean _coded) {
    coded = _coded;
    myKeyCode = _keyCode;
    if (!coded) {
      println("Unexpected invocation of Key(int, boolean) with boolean set to false. Continuing.");
    }
  }


  public boolean equals(Key k) {
    if (coded && k.coded) {
      return myKeyCode == k.myKeyCode;
    } else if (!coded && !k.coded) {
      return myKey == k.myKey;
    }
    return false;
  }

  public String toString() {
    return "Key : " + myKey + " KeyCode : " + myKeyCode + " Coded : " + coded;
  }
}
class NewGameState extends GameState{
  int backgroundColor = color(66);
  int verticalOffset = -50;

  public void setup() {
    leaderboard = new int[numPlayers];

    //Make the world
    world = new FWorld();
    world.setGrabbable(false);
    world.setGravity(0, 0);
    world.setEdges();
    world.setEdgesRestitution(0.5f);

    for (int i = 0; i < numPlayers; i++) {
      playerSides[i] = 3;
    }

    {
      //Make player 0
      PlayerController player = new PlayerController(numSides, 50.0f, world);
      player.setPosition((width / 2) - 100, (height / 2) + 100);
      player.setScale(scale.x, scale.y);
      player.setFill(new PVector(255, 0, 0));
      player.updateShape();
      player.tag = "Red";

      players[0] = player;
    }

    {
      //Make player 1
      PlayerController player = new PlayerController(numSides, 50.0f, world);
      player.setPosition((width / 2) + 100, (height / 2) + 100);
      player.setScale(scale.x, scale.y);
      player.setFill(new PVector(0, 0, 255));
      player.updateShape();
      player.tag = "Blue";

      players[1] = player;
    }
  }

  public void cleanup() {
  }

  public void pause() {
  }

  public void resume(GameState previousState) {
  }

  public void update() {
    world.step();

    for (int i = 0; i < numPlayers; i++) {
      PlayerController player = players[i];
      int sides = playerSides[i];
      player.setNumSides(sides);
      player.updateShape();
    }
  }

  public void draw() {
    background(backgroundColor);

    world.draw();
    textFont(omnesSem48);
    textAlign(CENTER);
    fill(255);
    text("Choose your sides!", width/2, (height/2 - 100)); 
    textFont(omnesSem29);
    text("Red: Left & right arrow keys to select sides", width/2, (height/2 + 200)); 
    text("Blue: W&D keys to select sides", width/2, (height/2 + 250)); 
    text("Space to play", width/2, (height/2 + 300)); 

    world.draw(getMainInstance());  
  }

  public void mouseDragged() {
  }

  public void mousePressed() {
  }

  public void mouseReleased() {
  }

  public void keyPressed() {
    //Key k = new Key(key, keyCode);

    //for (PlayerController player : players) {
    //  Input input = player.getInput();
    //  input.keyPressed(k);
    //  player.setInput(input);
    //}

    //HACK HACK HACK
    if (CODED == key) {
      if (LEFT == keyCode) {
        playerSides[0]--;
        playerSides[0] = constrain(playerSides[0], minSides, 999);
      } else if (RIGHT == keyCode) {
        playerSides[0]++;
        playerSides[0] = constrain(playerSides[0], minSides, 999);
      }
    } else {
      if ((ENTER == key) ||
          (RETURN == key) ||
          (' ' == key)
        ) {
        engineChangeState(new InterstitialState("Fight!", new PlayState(), interstitialLengthMs));
      } else if ('a' == key) {
        playerSides[1]--;
        playerSides[1] = constrain(playerSides[1], minSides, 999);
      } else if ('d' == key) {
        playerSides[1]++;
        playerSides[1] = constrain(playerSides[1], minSides, 999);
      }
    }
  }

  public void keyReleased() {
  }

}
/**
 * Math helper library for Processing
 * @author Eliot Lash
 */
static class PMath {

  public static PVector rotatePVector2DDebug(PVector _v, float theta) {
    PVector v = rotatePVector2D(_v, theta);
    println("_v " + _v + " theta " + theta + " v " +v);
    println("==========");
    return v;
  }

  /**
   * Based on code from http://www.shiffman.net/2011/02/03/rotate-a-vector-processing-js/
   */
  public static PVector rotatePVector2D(PVector _v, float theta) {
    PVector v = _v.get();
    v.x = _v.x * cos(theta) - _v.y * sin(theta);
    v.y = _v.x * sin(theta) + _v.y * cos(theta);

    return v;
  }
}
/**
 * Taken from the Spectrum Game Engine
 *
 * @author Eliot Lash
 * @copyright Copyright (c) 2010-2012 Eliot Lash
 */
class Path {

  ArrayList<PVector> points;
  float length = 0.0f;
  boolean lengthChanged = false;

  Path() {
    points = new ArrayList<PVector>();
  }

  /**
   * @return boolean True on successful add, false on failed add (path would exceed max length)
   */
  public boolean add(PVector point) {
    float newLength = newLength(point);
    //No need for max path length in this game
    //if (newLength <= maxPathLength) {
    points.add(point);
    length = newLength;
    return true;
    /*} else {
      return false;
    }*/
  }

  public float newLength(PVector point) {
    PVector lastPoint = lastPoint();
    if (null != lastPoint) {
      return length() + point.dist(lastPoint);
    } else {
      length = 0.0f;
      lengthChanged = false; //We just calculated that length is 0 since path length is 0 so we'll keep that calculation... hope this doesn't bite me in the ass later
      return length;
    }
  }

  /**
   * Reduces points array to distance float, if the length was flagged as changed, otherwise uses stored float value.
   * Fake lazy seq :(
   */
  public float length() {
    float newLength = 0.0f;
    if (lengthChanged) {
      PVector previousPoint = null;
      for (PVector currentPoint : points) {
        if (previousPoint == null) {
          previousPoint = currentPoint;
        }

        newLength += currentPoint.dist(previousPoint);
        previousPoint = currentPoint;
      }
    } else {
      newLength = length;
    }

    return newLength;
  }

  public PVector get(int i) {
    return points.get(i);
  }

  public void set(int i, PVector point) {
    points.set(i, point);
    lengthChanged = true;
  }

  public void clear() {
    points = new ArrayList<PVector>();
    length = 0.0f;
  }

  public void remove(int i) {
    points.remove(i);
    lengthChanged = true;
  }

  public int size() {
    return points.size();
  }

  public PVector lastPoint() {
    PVector lastPoint = null;
    if (lastElementIndex() >= 0) {
      lastPoint = points.get(lastElementIndex());
    }
    return lastPoint;
  }

  public int lastElementIndex() {
    return points.size() - 1;
  }

  //Draw path
  public void draw() {
    //println("draw begin @ " + points.size());
    pushStyle();
    stroke(color(255, 0, 0));
    strokeWeight(2);
    strokeJoin(ROUND);

    for (int i = 0; i < (points.size() - 2); i++) {
      //println("draw " + i + "/" + points.size());
      PVector currentVector = points.get(i);
      PVector nextVector = points.get(i+1);
      PVector thirdVector = points.get(i+2);
      line(currentVector.x, currentVector.y, nextVector.x, nextVector.y);
      //println("line(" + currentVector.x + "," + currentVector.y + "," + nextVector.x + "," + nextVector.y + ")");
      line(nextVector.x, nextVector.y, thirdVector.x, thirdVector.y);
    }
    popStyle();
  }

  public String toString() {
    return points.toString();
  }
}
class PlayState extends GameState {

  public void setup() {
    //Make the world
    world = new FWorld();
    world.setGrabbable(DEBUG); //Only allow mouse grabbing in debug mode
    world.setGravity(0, 0);
    world.setEdgesRestitution(0.5f);

    ellipseMode(RADIUS);

    {
      //Make player 0
      PlayerController player = new PlayerController(playerSides[0], 50.0f, world);
      player.setPosition((width / 2) - 100, (height / 2) - 100);
      player.setScale(scale.x, scale.y);
      player.setFill(new PVector(255, 0, 0));
      player.updateShape();
      player.tag = "Red";

      Input input = player.getInput();

      input.CENTER_KEY = new Key('/');
      input.UP_KEY = new Key(UP, true);
      input.DOWN_KEY = new Key(DOWN, true);
      input.LEFT_KEY = new Key(LEFT, true);
      input.RIGHT_KEY = new Key(RIGHT, true);
      input.SCALE_MOD_KEY = new Key(ALT, true);
      input.SCALE_X_NEGATIVE_KEY = new Key('j');
      input.SCALE_X_POSITIVE_KEY = new Key('l');
      input.SCALE_Y_NEGATIVE_KEY = new Key('k');
      input.SCALE_Y_POSITIVE_KEY = new Key('i');
      //input.useMouse = true;

      player.setInput(input);

      players[0] = player;
    }

    {
      //Make player 1
      PlayerController player = new PlayerController(playerSides[1], 50.0f, world);
      player.setPosition((width / 2) + 100, (height / 2) + 100);
      player.setScale(scale.x, scale.y);
      player.setFill(new PVector(0, 0, 255));
      player.updateShape();
      player.tag = "Blue";

      Input input = player.getInput();

      input.CENTER_KEY = new Key('c');
      input.UP_KEY = new Key('w');
      input.DOWN_KEY = new Key('s');
      input.LEFT_KEY = new Key('a');
      input.RIGHT_KEY = new Key('d');
      input.SCALE_MOD_KEY = new Key(SHIFT, true);
      input.SCALE_X_NEGATIVE_KEY = new Key('f');
      input.SCALE_X_POSITIVE_KEY = new Key('h');
      input.SCALE_Y_NEGATIVE_KEY = new Key('g');
      input.SCALE_Y_POSITIVE_KEY = new Key('t');

      player.setInput(input);

      players[1] = player;
    }

    //Obstacle
    PolygonController poly = new PolygonController(4, 25.0f, world);
    poly.setPosition(width / 2, height / 2);
    poly.setScale(scale.x, scale.y);
    poly.tag = "peg";
    //poly.setStatic(true);
    poly.setDensity(9999999999.0f);
    poly.updateShape();

    shapes.add(poly);

    //Make an enemy
    //EnemyController enemy = new EnemyController(3, 50.0, world);
    //enemy.setPosition(width/3, height/3);
    //enemy.updateShape();
    //enemies.add(enemy);
  }

  public void cleanup() {
  }

  public void pause() {
  }

  public void resume(GameState previousState) {
  }

  public void update() {
    for (PlayerController player : players) {
      Input input = player.getInput();

      /* //Rotation keys are now default, commenting out. 
      if (player.input.useMouse) {
        PVector mousePosition = gWorkVectorA;
        mousePosition.x = mouseX;
        mousePosition.y = mouseY;

        PVector pos = player.getPosition();
        PVector lookAt = PVector.sub(pos, mousePosition); //FIXME: Use workvector placeholders if this creates too much garbage
        lookAt.normalize();

        input.heading = lookAt;
      } else {
        //Process rotation keys if we're not using mouselook
        if (input.turnLeftKeyDown) {
        } else if (input.turnRightKeyDown) {
        }
      }*/

      //Set movement forces
      float xForce = 0;
      float yForce = 0;

      if (input.upKeyDown) {
        yForce = -movementForce;
      } else if (input.downKeyDown) {
        yForce = movementForce;
      }

      if (input.rightKeyDown) {
        input.rotation = radians(player.turnMagnitude);
      } else if (input.leftKeyDown) {
        input.rotation = radians(-player.turnMagnitude);
      } else {
        input.rotation = radians(0);
      }

      PVector playerScale = player.getScale();

      if (input.scaleXNegativeKeyDown) {
        playerScale.x -= scaleAdjustFactor;
      } else if (input.scaleXPositiveKeyDown) {
        playerScale.x += scaleAdjustFactor;
      }

      if (input.scaleYNegativeKeyDown) {
        println(player+"scaleY-");
        playerScale.y -= scaleAdjustFactor;
      } else if (input.scaleYPositiveKeyDown) {
        playerScale.y += scaleAdjustFactor;
      }

      player.setScale(playerScale.x, playerScale.y);

      if (input.centerKeyDown) {
        player.resetScale();
      }

      player.updateShape();

      player.setRelativeForce(xForce, yForce);
      player.update();
    }

    if (!disableEnemyUpdate) {
      for (EnemyController enemy : enemies) {
        enemy.update();
      }
    }

    //Check for win condition
    for (int i = 0; i < numPlayers; i++) {
      PlayerController player = players[i];
      PVector position = player.getPosition();
      if (
          (position.x > width) ||
          (position.x < 0) ||
          (position.y > height) ||
          (position.y < 0)
         ) {
        println(width + "," + height);
        println(position);
        leaderboard[1 - i]++; //the other player gets a point, only works for 2 players FIXME

        boolean matchWinner = false;
        int matchWinnerNum = -1;
        for (int j = 0 ; j < numPlayers; j++) {
          int score = leaderboard[j];
          if (score >= maxScore) {
            matchWinner = true;
            matchWinnerNum = j;
          }
        }

        if (matchWinner) {
          engineChangeState(new InterstitialState(players[matchWinnerNum] + " wins game!", new NewGameState()));
        } else {
          engineChangeState(new InterstitialState(player + " was knocked out!", new InterstitialState("Fight!", new PlayState(), interstitialLengthMs), interstitialLengthMs));
        }
      }
    }

    world.step();
  }

  public void draw() {
    background(66);

    world.draw(getMainInstance());  

    for (PlayerController player : players) {
      player.draw();
    }

    //Debug draw
    if (DEBUG) {
      for (EnemyController enemy : enemies) {
        enemy.drawPath();
      }
    }
  }

  public void mouseDragged() {
  }

  public void mousePressed() {
  }

  public void mouseReleased() {
  }

  public void keyPressed() {
    Key k = new Key(key, keyCode);
    //println("Key pressed - " + k);

    for (PlayerController player : players) {
      Input input = player.getInput();
      input.keyPressed(k);
      player.setInput(input);
    }

    if (CODED == key) {
    } else {
      //DEBUG keys
      if (DEBUG) {
        if ('p' == key) {
            disableEnemyUpdate = !disableEnemyUpdate;
            println(disableEnemyUpdate);
        } else if ('=' == key) {
          numSides++;
          println("numSides = " + numSides);
        } else if ('-' == key) {
          numSides--;
          numSides = constrain(numSides, minSides, 999);
          println("numSides = " + numSides);
        } else if ('+' == key) { //Shift +
          scale.x += scaleAdjustFactor; 
          println("scale.x = " + scale.x);
        } else if ('_' == key) { //Shift =
          scale.x -= scaleAdjustFactor; 
          println("scale.x = " + scale.x);
        } else if ('0' == key) {
          PolygonController poly = new PolygonController(numSides, 50.0f, world);
          poly.setPosition(mouseX, mouseY);
          poly.setWorld(world);
          poly.setScale(scale.x, scale.y);
          poly.updateShape();
          shapes.add(poly);
        } else if (key == BACKSPACE) {
          FBody hovered = world.getBody(mouseX, mouseY);
          if ( hovered != null &&
               hovered.isStatic() == false ) {
            //Destroy the body and PolygonController
            world.remove(hovered);

            PolygonController hoveredController = null;
            for (PolygonController p : shapes) {
              if (hovered == p.getPoly()) {
                hoveredController = p;
                break;
              }
            }
            if (null != hoveredController) {
              shapes.remove(hoveredController);
            }
          }
        }
      } 
    }
  }

  public void keyReleased() {
    Key k = new Key(key, keyCode);

    for (PlayerController player : players) {
      Input input = player.getInput();
      input.keyReleased(k);
      player.setInput(input);
    }
  }
}
class PlayerController extends PolygonController {
  PVector mousePosition;
  Input input;
  float turnMagnitude = 135;

  PlayerController(int numSides, float radius) {
    super(numSides, radius);
    construct();
  }
  
  PlayerController(int numSides, float radius, FWorld world) {
    super(numSides, radius, world);
    construct();
  }

  public void construct() {
    mousePosition = new PVector();
    invincible = false;
    input = new Input();
  }

  public Input getInput() {
    //If better encapsulation / gurantee of non-mutation is needed, change this call to return a copy, eg:
    //return new Input(input);
    return input;
  }
  
  public void setInput(Input newInput) {
    input = newInput;
  }

  public void update() {
    super.update();

    if (null != poly) {

      //Calculate & apply rotation
      getHeading(); //update this.heading

      poly.setAngularVelocity(input.rotation);

      //Collision detection
      ArrayList<FContact> contacts = poly.getContacts();
      for (FContact c : contacts) {
        EnemyController enemy = findEnemy(c.getBody2());
        if (null == enemy) {
          //That wasn't it, let's try the first body...
          enemy = findEnemy(c.getBody1());
        }
        if (null != enemy) {
          line(c.getBody1().getX(), c.getBody1().getY(), c.getBody2().getX(), c.getBody2().getY());
        }
      }
    }
  }

  public void draw() {
    getPosition();

    pushStyle();
    fill(0);
    ellipse(position.x, position.y, centerDotRadius, centerDotRadius);
    popStyle();

    if (DEBUG) {
      getHeading();
      getPosition();
      PVector headingNormal = getHeadingNormal();
      pushStyle();
      strokeWeight(2);
      pushStyle();
      stroke(0, 255, 0);
      line(position.x, position.y, position.x + heading.x * 50, position.y + heading.y * 50);
      popStyle();
      pushStyle();
      stroke(255, 0, 0);
      line(position.x, position.y, position.x + headingNormal.x * 50, position.y + headingNormal.y * 50);
      popStyle();
      popStyle();
      //println("heading: " + heading + " headingNormal: " + headingNormal);
    }
  }
}
class Polygon extends FPoly {
    int sideCount = 0;
  
    public void setSideCount(int n) {
      sideCount = n;
    }
    
    public int getSideCount() {
      return sideCount;
    }
  
    public void draw(PGraphics applet) {
      super.draw(applet);
    }
}
class PolygonController {
  Polygon poly;
  FWorld world;
  int numSides;
  float radius;
  PVector defaultScale;
  PVector scale;
  PVector position;
  PVector heading;
  float turnSpeed = 7;
  boolean invincible = true;
  int health = 50;
  PVector fillColor;
  String tag;
  static final float centerDotRadius = 3;
  boolean isStatic = false;
  float density = 10;

  /** Time delta since last update in milliseconds */
  float deltaTime = 0; 
  /** Time of last update in milliseconds */
  float lastUpdate = 0;

  static final float polyCalcRotationOffset = -1.5707964f; //Rotation offset when creating the polygon so it 'points' at the cursor. -1.5707964 rads = -90 degrees.
  static final float scaleLimit = 0.02f; //The distance the scale on any axis is allowed to be from 0, if it gets too close to 0 it causes the physics engine to glitch

  //Re-using some PVector objects to reduce garbage during calcs in a tight loop
  PVector workVectorA;
  PVector workVectorB;
  PVector workVectorC;
  
  PolygonController(int numSides, float radius) {
    construct(numSides, radius, null);
  }
  
  PolygonController(int numSides, float radius, FWorld world) {
    construct(numSides, radius, world);
  }
  
  public void construct(int numSides, float radius, FWorld world) {
    defaultScale = new PVector(1,1);
    scale = new PVector(1,1);
    position = new PVector();
    heading = new PVector();
    fillColor = new PVector(120, 30, 90);

    workVectorA = new PVector();
    workVectorB = new PVector();
    workVectorC = new PVector();

    this.numSides = numSides;
    this.radius = radius;
    
    if (null != world) {
      setWorld(world);
    }
  }
  
  public void setPosition(float x, float y) {
    position.x = x;
    position.y = y;

    if (null != poly) {
      poly.setPosition(x, y);
    }
  }

  public PVector getPosition() {
    if (null != poly) {
      position.x = poly.getX();
      position.y = poly.getY();
    }
    return position;
  }

  public PVector getHeading() {
    if (null != poly) {
      float rotation = poly.getRotation();
      heading.x = cos(rotation);
      heading.y = sin(rotation);
    }
    return heading;
  }

  public PVector getHeadingNormal() {
    return PMath.rotatePVector2D(heading, radians(90));
  }

  public void setNumSides(int n) {
    numSides = n;
  }

  public PVector getScale() {
    return scale.get();
  }
  public void resetScale() {
    setScale(defaultScale.x, defaultScale.y);
  }

  public void setScale(float u) {
    setScale(u, u);
  }

  public void setScale(float x, float y) {
    float aspectRatio = x / y;
    //println("scaleX: " + x + " scaleY: " + y + " aspectRatio: " + aspectRatio);
    if (
        isFurtherThan(aspectRatio, 0, scaleLimit)
        && (aspectRatio < 33)
        && (aspectRatio > -33)
    ) {
      //println ("scale good.");
      scale.x = x;
      scale.y = y;
    }
  }

  public boolean isFurtherThan(float a, float b, float range) {
    return (abs(a - b) > range);
  }

  public void setForce(float fx, float fy) {
    poly.setForce(fx, fy);
  }

  public void setRelativeForce(float fx, float fy) {
    PVector headingNormal = getHeadingNormal();
    getHeading();
    //Apply transformations
    //[a c
    // b d]
    workVectorA.x = heading.x; //a
    workVectorA.y = heading.y; //b

    workVectorB.x = headingNormal.x; //c
    workVectorB.y = headingNormal.y; //d

    //x(a,b) + y(c,d)
    workVectorA.mult(fx);
    workVectorB.mult(fy);
    PVector force = PVector.add(workVectorA, workVectorB); //If this adds too much garbage, try instance .add() on another work vector

    poly.setForce(force.x, force.y);
  }
  
  public void setWorld(FWorld world) {
    this.world = world;
  }

  public void setStatic(boolean s) {
    isStatic = s;
  }
  
  public void addToWorld() {
    world.add(poly);
    //println("added " + this + " to world.");
  }

  //Workaround for color passing to fisica being buggy
  public void setFill(PVector newColor) {
    fillColor = newColor;
  }
  
  public FPoly getPoly() {
    return poly;
  }

  public boolean isInvincible() {
    return invincible;
  }
  
  public void setDensity(float _density) {
    density = _density;
  }
  /**
   * You MUST call updateShape before the new PolygonController is usable.
   */
  public void updateShape() {
    int sideCount = numSides;
    float[][] vertices = new float[sideCount][2]; //[n][0] = x, [n][1] = y

    //Create an array of cartesian coords for a regular polygon of n sides
    //See http://stackoverflow.com/questions/7198144/how-to-draw-a-n-sided-regular-polygon-in-cartesian-coordinates
    for (int vertexNum = 0; vertexNum < sideCount; vertexNum++) {
      vertices[vertexNum][0] = radius * cos(2*PI*vertexNum/sideCount + polyCalcRotationOffset);
      vertices[vertexNum][1] = radius * sin(2*PI*vertexNum/sideCount + polyCalcRotationOffset);
    }
    
    Polygon oldPoly = poly;

    //Build new physics body
    poly = new Polygon();
    poly.setBullet(true);
    poly.setStrokeWeight(3);
    poly.setFill(fillColor.x, fillColor.y, fillColor.z);
    poly.setDensity(density);
    poly.setRestitution(0.5f);
    poly.setStatic(isStatic);
    if (null != oldPoly) {
      //Copy the physics properties that are likely to be different from the old body.
      PVector oldPosition = workVectorC;
      oldPosition.x = oldPoly.getX();
      oldPosition.y = oldPoly.getY();
      
      if ((oldPosition.x == 0) && (oldPosition.y == 0)) {
        oldPosition = position;
      }

      poly.setPosition(oldPosition.x, oldPosition.y);
      //println("oldPosition: " + oldPoly.getX() + ", " + oldPoly.getY());
      poly.setRotation(oldPoly.getRotation());
      poly.setAngularVelocity(oldPoly.getAngularVelocity());
      poly.setForce(oldPoly.getForceX(), oldPoly.getForceY());
      poly.setVelocity(oldPoly.getVelocityX(), oldPoly.getVelocityY());
    } else {
      //No old body, use caller supplied position or default
      poly.setPosition(position.x, position.y);
    }
    poly.setSideCount(numSides);

    for (int vertexNum = 0; vertexNum < sideCount; vertexNum++) {
      PVector coords;
      float x = vertices[vertexNum][0]; 
      float y = vertices[vertexNum][1];

      //Apply transformations
      //[a c
      // b d]
      workVectorA.x = scale.x; //a
      workVectorA.y = 0; //b

      workVectorB.x = 0; //c
      workVectorB.y = scale.y; //d

      //x(a,b) + y(c,d)
      workVectorA.mult(x);
      workVectorB.mult(y);
      coords = PVector.add(workVectorA, workVectorB); //If this adds too much garbage, try instance .add() on another work vector

      poly.vertex(coords.x, coords.y);
    }
    
    if (null != world) {
      world.remove(oldPoly);
      addToWorld();
    }
  }

  public void update() {
    float ms = millis();
    deltaTime = ms - lastUpdate;
    lastUpdate = ms;
  }

  public void draw() {
  }

  public String toString() {
    if (null != tag) {
      return tag;
    } else {
      return super.toString();
    }
  }
}
/**
 * Taken from the Spectrum Game Engine
 *
 * @author Eliot Lash
 * @copyright Copyright (c) 2010-2012 Eliot Lash
 */
class Stack extends ArrayList {

  /**
   * Push an object onto the stack
   */
  public void push(Object o) {
    this.add(o);
  }

  /**
   * Pop an object off the stack
   * @return Object The Object on top of the stack, or null if stack is empty. 
   */
  public Object pop() {
    int lastElement = this.lastElementIndex();
    Object returnObject = null;
    if (lastElement >= 0) {
      returnObject = this.get(lastElement);
      this.remove(lastElement);
    }
    return returnObject;
  }

  /**
   * Peek at the object on the top of the stack
   * @return Object The Object on top of the stack, or null if stack is empty. 
   */
  public Object peek() {
    int lastElement = this.lastElementIndex();
    Object returnObject = null;
    if (lastElement >= 0) {
      returnObject = this.get(lastElement);
    }
    return returnObject;
  }

  /**
   * Helper to get the array index of the last element on the stack
   * @return int Array index of last element. Negative values indicate that the stack is empty.
   */
  private int lastElementIndex() {
    return (this.size() - 1);
  }

}
  static public void main(String args[]) {
    PApplet.main(new String[] { "--present", "--bgcolor=#666666", "--hide-stop", "Sumogons" });
  }
}
